exception Exn of string

type ptr = int ref
and bc =
  | Acc of int
  | Push
  | Pop of int
  | Assign of int
  | Envacc of int
  | Pushretaddr of ptr
  | Apply of int
  | Appterm of int * int
  | Return of int
  | Restart
  | Grab of int
  | Closure of int * ptr
  | Closurerec of int * int * ptr * ptr array
  | Offsetclosure of int
  | Getglobal of int
  | Getglobalfield of int * int
  | Setglobal of int
  | Atom of int
  | Makeblock of int * int
  | Makefloatblock of int
  | Getfield of int
  | Getfloatfield of int
  | Setfield of int
  | Setfloatfield of int
  | Vectlength
  | Getvectitem
  | Setvectitem
  | Getstringchar
  | Setstringchar
  | Branch of ptr
  | Branchif of ptr
  | Branchifnot of ptr
  | Switch of int * ptr array
  | Boolnot
  | Pushtrap of ptr
  | Poptrap
  | Raise
  | Checksignals
  | Ccall of int * int
  | Const of int
  | Negint
  | Addint
  | Subint
  | Mulint
  | Divint
  | Modint
  | Andint
  | Orint
  | Xorint
  | Lslint
  | Lsrint
  | Asrint
  | Eq
  | Neq
  | Ltint
  | Leint
  | Gtint
  | Geint
  | Offsetint of int
  | Offsetref of int
  | Isint
  | Getmethod
  | Beq of int * ptr
  | Bneq of int * ptr
  | Blint of int * ptr
  | Bleint of int * ptr
  | Bgtint of int * ptr
  | Bgeint of int * ptr
  | Ultint
  | Ugeint
  | Bultint of int * ptr
  | Bugeint of int * ptr
  | Getpubmet of int * int
  | Getdynmet
  | Stop
  | Event
  | Break
  | Nop
;;

let print_instr oc instr =
  match instr with
    | Acc n             -> Printf.fprintf oc "ACC %d" n
    | Push              -> Printf.fprintf oc "PUSH"
    | Pop n             -> Printf.fprintf oc "POP %d" n
    | Assign n          -> Printf.fprintf oc "ASSIGN %d" n
    | Envacc n          -> Printf.fprintf oc "ENVACC %d" n
    | Pushretaddr ptr   -> Printf.fprintf oc "PUSHRETADDR %d" !ptr
    | Apply n           -> Printf.fprintf oc "APPLY %d" n
    | Appterm (n,s)     -> Printf.fprintf oc "APPTERM %d %d" n s
    | Return n          -> Printf.fprintf oc "RETURN %d" n
    | Restart           -> Printf.fprintf oc "RESTART"
    | Grab n            -> Printf.fprintf oc "GRAB %d" n
    | Closure (n,ptr)   -> Printf.fprintf oc "CLOSURE %d %d" n !ptr
    | Closurerec (f,v,{ contents = o },t) ->
	Printf.fprintf oc "CLOSUREREC %d %d %d [" f v o;
	Array.iter (fun p -> Printf.fprintf oc " %d " !p) t;
	Printf.fprintf oc "]";
    | Offsetclosure n       -> Printf.fprintf oc "OFFSETCLOSURE %d" n
    | Getglobal n           -> Printf.fprintf oc "GETGLOBAL %d" n
    | Getglobalfield (n,p)  -> Printf.fprintf oc "GETGLOBALFIELD %d %d"n p
    | Setglobal n       -> Printf.fprintf oc "SETGLOBAL %d" n
    | Atom n            -> Printf.fprintf oc "ATOM %d" n
    | Makeblock (n,t)   -> Printf.fprintf oc "MAKEBLOCK %d %d" n t
    | Makefloatblock n  -> Printf.fprintf oc "MAKEFLOATBLOCK %d" n
    | Getfield n        -> Printf.fprintf oc "GETFIELD %d" n
    | Getfloatfield n   -> Printf.fprintf oc "GETFLOATFIELD %d" n
    | Setfield n        -> Printf.fprintf oc "SETFIELD %d" n
    | Setfloatfield n   -> Printf.fprintf oc "SETFLOATFIELD %d" n
    | Vectlength        -> Printf.fprintf oc "VECTLENGTH"
    | Getvectitem       -> Printf.fprintf oc "GETVECTITEM"
    | Setvectitem       -> Printf.fprintf oc "SETVECTITEM"
    | Getstringchar     -> Printf.fprintf oc "GETSTRINGCHAR"
    | Setstringchar     -> Printf.fprintf oc "SETSTRINGCHAR"
    | Branch ptr        -> Printf.fprintf oc "BRANCH %d" !ptr
    | Branchif ptr      -> Printf.fprintf oc "BRANCHIF %d" !ptr
    | Branchifnot ptr   -> Printf.fprintf oc "BRANCHIFNOT %d" !ptr
    | Switch (n, tab) ->
	let size_tag = n lsr 16 in
	let size_long = n land 0xFFFF in
	  Printf.fprintf oc "SWITCH %d %d [" size_tag size_long;
	  Array.iter (fun ptr -> Printf.fprintf oc " %d " !ptr) tab;
	  Printf.fprintf oc "]";
    | Boolnot           -> Printf.fprintf oc "BOOLNOT"
    | Pushtrap ptr      -> Printf.fprintf oc "PUSHTRAP %d" !ptr
    | Poptrap           -> Printf.fprintf oc "POPTRAP"
    | Raise             -> Printf.fprintf oc "RAISE"
    | Checksignals      -> Printf.fprintf oc "CHECKSIGNALS"
    | Ccall (n, ind)    -> Printf.fprintf oc "CCALL %d %d" n ind
    | Const n           -> Printf.fprintf oc "CONST %d" n
    | Negint            -> Printf.fprintf oc "NEGINT"
    | Addint            -> Printf.fprintf oc "ADDINT"
    | Subint            -> Printf.fprintf oc "SUBINT"
    | Mulint            -> Printf.fprintf oc "MULINT"
    | Divint            -> Printf.fprintf oc "DIVINT"
    | Modint            -> Printf.fprintf oc "MODINT"
    | Andint            -> Printf.fprintf oc "ANDINT"
    | Orint             -> Printf.fprintf oc "ORINT"
    | Xorint            -> Printf.fprintf oc "XORINT"
    | Lslint            -> Printf.fprintf oc "LSLINT"
    | Lsrint            -> Printf.fprintf oc "LSRINT"
    | Asrint            -> Printf.fprintf oc "ASRINT"
    | Eq                -> Printf.fprintf oc "EQ"
    | Neq               -> Printf.fprintf oc "NEQ"
    | Ltint             -> Printf.fprintf oc "LTINT"
    | Leint             -> Printf.fprintf oc "LEINT"
    | Gtint             -> Printf.fprintf oc "GTINT"
    | Geint             -> Printf.fprintf oc "GEINT"
    | Offsetint ofs     -> Printf.fprintf oc "OFFSETINT %d" ofs
    | Offsetref ofs     -> Printf.fprintf oc "OFFSETREF %d" ofs
    | Isint             -> Printf.fprintf oc "ISINT"
    | Getmethod         -> Printf.fprintf oc "GETMETHOD"
    | Beq (v,ptr)       -> Printf.fprintf oc "BEQ %d %d" v !ptr
    | Bneq (v,ptr)      -> Printf.fprintf oc "BNEQ %d %d" v !ptr
    | Blint (v,ptr)     -> Printf.fprintf oc "BLINT %d %d" v !ptr
    | Bleint (v,ptr)    -> Printf.fprintf oc "BLEINT %d %d" v !ptr
    | Bgtint (v,ptr)    -> Printf.fprintf oc "BGTINT %d %d" v !ptr
    | Bgeint (v,ptr)    -> Printf.fprintf oc "BGEINT %d %d" v !ptr
    | Ultint            -> Printf.fprintf oc "ULTINT"
    | Ugeint            -> Printf.fprintf oc "UGEINT"
    | Bultint (v,ptr)   -> Printf.fprintf oc "BULTINT %d %d" v !ptr
    | Bugeint (v,ptr)   -> Printf.fprintf oc "BUGEINT %d %d" v !ptr
    | Getpubmet (v,cch) -> Printf.fprintf oc "GETPUBMET %d %d" v cch
    | Getdynmet         -> Printf.fprintf oc "GETDYNMET"
    | Stop              -> Printf.fprintf oc "STOP"
    | Event             -> Printf.fprintf oc "EVENT"
    | Break             -> Printf.fprintf oc "BREAK"
    | Nop               -> Printf.fprintf oc "NOP"
;;
